//aabb.cpp
#include <assert.h>
#include <d3d9.h>
#include <d3dx9.h>
#include "ray.h"
#include "aabb.h"

//AABB 浹 ƾ Ʒ Ʈ ߴ.
//http://www.nuclex.org/articles/static-intersection-test-between-aabb-and-aabb 

const float AABB::INFINITY = FLT_MAX;
const float AABB::EPSILON = 0.001f;

//------------------------------------------------------------------------------------------------
AABB::AABB() 
{
	m_min = D3DXVECTOR3( FLT_MAX, FLT_MAX, FLT_MAX ); 
	m_max = D3DXVECTOR3( -FLT_MAX, -FLT_MAX, -FLT_MAX );
}

//------------------------------------------------------------------------------------------------
AABB::~AABB()
{
}
//------------------------------------------------------------------------------------------------
bool AABB::Equals( float lhs, float rhs )
{
	return fabs( lhs - rhs ) < EPSILON ? true : false;
}
//------------------------------------------------------------------------------------------------
void AABB::GetMinMax( D3DXVECTOR3& _min, D3DXVECTOR3 _max )
{
	_min = m_min;
	_max = m_max;
}
//------------------------------------------------------------------------------------------------
void AABB::SetMinMax( D3DXVECTOR3& _min, D3DXVECTOR3 _max )
{
	m_min = _min;
	m_max = _max;
}
//------------------------------------------------------------------------------------------------
bool AABB::TestPoint( const D3DXVECTOR3& pos )
{
	if( pos.x >= m_min.x && pos.x < m_max.x )
	{
		if( pos.y >= m_min.y && pos.y < m_max.y )
		{
			if( pos.z >= m_min.z && pos.z < m_max.z )		
			{
				return true;
			}
		}
	}

	return false;
}
//------------------------------------------------------------------------------------------------
bool AABB::TestAABB( AABB& box )
{
	D3DXVECTOR3 _min, _max;
	box.GetMinMax( _min, _max );

	return  (   (m_min.x < _max.x) && (m_max.x > _min.x) && 
				(m_min.y < _max.y) && (m_max.y > _min.y) && 
				(m_min.z < _max.z) && (m_max.z > _min.z)  );
}
//------------------------------------------------------------------------------------------------
bool AABB::TestSphere( const D3DXVECTOR3& sphereCenter, float sphereRadius ) 
{
	float totalDistance = 0;

	// Accumulate the distance of the sphere's center on each axis
	for(int i = 0; i < 3; ++i) 
	{
		// If the sphere's center is outside the aabb, we've got distance on this axis
		if(sphereCenter[i] < m_min[i]) 
		{
			float borderDistance = m_min[i] - sphereCenter[i];
			totalDistance += borderDistance * borderDistance;
		} 
		else if(sphereCenter[i] > m_max[i]) 
		{
			float borderDistance = sphereCenter[i] - m_max[i];
			totalDistance += borderDistance * borderDistance;
		}
		// Otherwise the sphere's center is within the box on this axis, so the
		// distance will be 0 and we do not need to accumulate anything at all
	}

	// If the distance to the box is lower than the sphere's radius, both are overlapping
	return (totalDistance <= (sphereRadius * sphereRadius));
}
//------------------------------------------------------------------------------------------------
bool AABB::TestRay(const Ray &ray)
{
	// Compute the ray delta
	D3DXVECTOR3 rayDelta = ray.GetDir() * ray.GetLength();

	// First we check to see if the origin of the ray is inside the AABB.  If it is, the ray intersects the AABB so 
	// we'll return true.  We start by assuming the ray origin is nside the AABB
	bool inside = true;

	// This stores the distance from either the min or max (x,y,z) to the ray's
	// origin (x,y,z) respectively, divided by the length of the ray.  The largest
	// value has the delta time of a possible intersection.
	float xt, yt, zt;
	
	// Test the X component of the ray's origin to see if we are inside or not
	if(ray.GetOrg().x < m_min.x)
	{
		xt = m_min.x - ray.GetOrg().x;
		
		if(xt > rayDelta.x) // If the ray is moving away from the AABB, there is no intersection 
			return false;
		
		xt /= rayDelta.x; 
		inside = false;
	} 
	else if(ray.GetOrg().x > m_max.x)
	{
		xt = m_max.x - ray.GetOrg().x;
		
		if(xt < rayDelta.x) // If the ray is moving away from the AABB, there is no intersection 
			return false;
		
		xt /= rayDelta.x;
		inside = false;
	} 
	else
	{
		// Later on we use the "xt", "yt", and "zt" variables to determine which plane (either
		// xy, xz, or yz) we may collide with.  Since the x component of the ray
		// origin is in between, the AABB's left and right side (which reside in the yz plane), 
		// we know we don't have to test those sides so we set this to a negative value.
		xt = -1.0f; 
	}

	// Test the X component of the ray's origin to see if we are inside or not
	if(ray.GetOrg().y < m_min.y)
	{
		yt = m_min.y - ray.GetOrg().y;
		
		if(yt > rayDelta.y) // If the ray is moving away from the AABB, there is no intersection 
			return false;

		yt /= rayDelta.y;
		inside = false;
	} 
	else if(ray.GetOrg().y > m_max.y)
	{
		yt = m_max.y - ray.GetOrg().y;
		
		if(yt < rayDelta.y) // If the ray is moving away from the AABB, there is no intersection 
			return false;
		
		yt /= rayDelta.y;
		inside = false;
	} 
	else
	{
		// Later on we use the "xt", "yt", and "zt" variables to determine which plane (either
		// xy, xz, or yz) we may collide with.  Since the y component of the ray
		// origin is in between, the AABB's top and bottom side (which reside in the xz plane), 
		// we know we don't have to test those sides so we set this to a negative value.
		yt = -1.0f;
	}

	if(ray.GetOrg().z < m_min.z)
	{
		zt = m_min.z - ray.GetOrg().z;
		
		if(zt > rayDelta.z) // If the ray is moving away from the AABB, there is no intersection 
			return false;
		
		zt /= rayDelta.z;
		inside = false;
	} 
	else if(ray.GetOrg().z > m_max.z)
	{
		zt = m_max.z - ray.GetOrg().z;
		
		if(zt < rayDelta.z) // If the ray is moving away from the AABB, there is no intersection 
			return false;
		
		zt /= rayDelta.z;
		inside = false;
	} 
	else
	{
		// Later on we use the "xt", "yt", and "zt" variables to determine which plane (either
		// xy, xz, or yz) we may collide with.  Since the z component of the ray
		// origin is in between, the AABB's front and back side (which reside in the xy plane), 
		// we know we don't have to test those sides so we set this to a negative value.
		zt = -1.0f;
	}

	// If the origin inside the AABB
	if(inside)
		return true; // The ray intersects the AABB
	
	// Otherwise we have some checking to do...

	// We want to test the AABB planes with largest value out of xt, yt, and zt.  So
	// first we determine which value is the largest.

	float t = xt;
	
	if(yt > t)
		t = yt;
	
	if(zt > t)
		t = zt;

	// **NOTE** Normally comparing two floating point numbers won't necessarily work, however,
	//			since we set to explicitly to equal either xt, yt, or zt above, the equality test
	//			will pass

	if(t == xt) // If the ray intersects with the AABB's YZ plane
	{
		// Compute intersection values
		float y = ray.GetOrg().y + rayDelta.y * t;
		float z = ray.GetOrg().z + rayDelta.z * t;
	
		// Test to see if collision takes place within the bounds of the AABB
		if(y < m_min.y || y > m_max.y)
			return false;
		else if(z < m_min.z || z > m_max.z)
			return false;
	}
	else if(t == yt) // Intersects with the XZ plane
	{
		// Compute intersection values
		float x = ray.GetOrg().x + rayDelta.x * t;
		float z = ray.GetOrg().z + rayDelta.z * t;

		// Test to see if collision takes place within the bounds of the AABB
		if(x < m_min.x || x > m_max.x)
			return false;
		else if(z < m_min.z || z > m_max.z) 
			return false;
	}
	else // Intersects with the XY plane
	{
		assert(t == zt);

		// Compute intersection values
		float x = ray.GetOrg().x + rayDelta.x * t;
		float y = ray.GetOrg().y + rayDelta.y * t;

		// Test to see if collision takes place within the bounds of the AABB
		if(x < m_min.x || x > m_max.x)
			return false;
		else if(y < m_min.y || y > m_max.y)
			return false;
	}

	// The ray intersects the AABB
	return true;
}
